from pydantic_settings import SettingsConfigDict

try:
    from aiohttp import BasicAuth, ClientSession
except ImportError:  # pragma: no cover
    BasicAuth = None
    ClientSession = None

from typing import Optional, Type

from pydantic import SecretStr, constr

from .base import BaseBackend, BaseNotification, BaseResponse

KeyTypeStr = constr(pattern=r"^\+?[1-9]\d{1,14}$")


class TwilioSMSResponse(BaseResponse):
    success: bool
    status_code: int = None
    request_id: str = None
    message: str = None


class TwilioSMSNotification(BaseNotification):
    model_config = SettingsConfigDict(
        arbitrary_types_allowed=True,
        env_prefix="notifications_twilio_sms_",
    )

    from_number: str
    to_number: KeyTypeStr
    body: str
    base_url: str
    account_sid: SecretStr
    auth_token: SecretStr
    fail_silently: bool = False
    response: Optional[TwilioSMSResponse] = None

    @property
    def backend(self) -> Type["TwilioSMSBackend"]:
        return TwilioSMSBackend


class TwilioSMSBackend(BaseBackend):
    options: TwilioSMSNotification
    connection = None

    def __init__(self, options: TwilioSMSNotification) -> None:
        super().__init__(options)

        if not ClientSession:  # pragma: no cover
            raise ModuleNotFoundError("python library aiohttp required")

    async def send(self) -> bool:
        auth = BasicAuth(
            login=self.options.account_sid.get_secret_value(),
            password=self.options.auth_token.get_secret_value(),
        )
        async with ClientSession(auth=auth) as connection:
            try:
                data = {
                    "From": self.options.from_number,
                    "To": self.options.to_number,
                    "Body": self.options.body,
                }
                post_data = await connection.post(
                    f"{self.options.base_url}/2010-04-01/Accounts/{self.options.account_sid}/Messages.json",
                    data=data,
                )
                body = await post_data.json()
                # Twilio uses 201 for success
                if post_data.status != 201:
                    self.options.response = TwilioSMSResponse(
                        status_code=post_data.status,
                        success=False,
                        request_id=post_data.headers["Twilio-Request-Id"],
                        message=body["message"],
                    )
                    return False
                else:
                    self.options.response = TwilioSMSResponse(
                        status_code=post_data.status,
                        success=True,
                        request_id=post_data.headers["Twilio-Request-Id"],
                        message=body["message"],
                    )
                    return True
            except Exception:
                self.options.response = TwilioSMSResponse(success=False)
                if self.options.fail_silently:
                    return False
                raise



from email import message_from_bytes, message_from_string
from email.message import Message
from typing import Any, Callable

from pydantic_core import core_schema


class Email(Message):
    @classmethod
    def __get_validators__(cls):
        yield cls._validate

    @classmethod
    def _validate(cls, v) -> Message:
        try:
            if isinstance(v, str):
                return message_from_string(v)
            elif isinstance(v, bytes):
                return message_from_bytes(v)
        except Exception as e:
            raise ValueError("invalid format.") from e

        if not isinstance(v, Message):
            raise ValueError("invalid format.")

        return v

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Callable[[Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        def validate_from_str(value: str) -> Message:
            result = message_from_string(value)
            return result

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_instance_schema(Message),
                    from_str_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: instance.as_string()
            ),
        )



from typing import Optional, Type

from pydantic import BaseModel
from pydantic_settings import BaseSettings


class BaseResponse(BaseModel):
    success: bool


class BaseNotification(BaseSettings):
    response: Optional[BaseResponse] = None

    @property
    def backend(self) -> Type["BaseBackend"]:
        raise NotImplementedError()


class BaseBackend:
    options: BaseNotification

    def __init__(self, options: BaseNotification) -> None:
        self.options = options

    async def __aenter__(self):
        try:
            await self.open()
        except Exception:
            await self.close()
            raise
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()

    async def open(self):
        pass

    async def close(self):
        pass

    async def send(self) -> bool:
        raise NotImplementedError()



from typing import Optional, Type

from pydantic_settings import SettingsConfigDict

try:
    from aiosmtplib import SMTP
except ImportError:  # pragma: no cover
    SMTP = None

from pydantic import SecretStr

from accentnotifications.types import Email

from .base import BaseBackend, BaseNotification, BaseResponse


class SmtpResponse(BaseResponse):
    success: bool


class SmtpNotification(BaseNotification):
    model_config = SettingsConfigDict(
        arbitrary_types_allowed=True,
        env_prefix="notifications_smtp_",
        # json_encoders={Message: lambda v: v.as_string()}
    )

    host: str
    port: int
    username: Optional[SecretStr] = None
    password: Optional[SecretStr] = None
    use_tls: bool = False
    starttls: bool = False
    timeout: Optional[int] = None
    fail_silently: bool = False
    email: Email
    response: Optional[SmtpResponse] = None

    @property
    def backend(self) -> Type["SmtpBackend"]:
        return SmtpBackend


class SmtpBackend(BaseBackend):
    options: SmtpNotification
    connection = None

    def __init__(self, options: BaseNotification) -> None:
        super().__init__(options)

        if not SMTP:  # pragma: no cover
            raise ModuleNotFoundError("python library aiosmtplib required")

    async def open(self):
        params = {
            "hostname": self.options.host,
            "port": self.options.port,
            "use_tls": self.options.use_tls and not self.options.starttls,
        }
        if self.options.timeout is not None:
            params["timeout"] = self.options.timeout

        try:
            self.connection = SMTP(**params)

            await self.connection.connect()

            if self.options.starttls:
                await self.connection.starttls()

            if self.options.username and self.options.password:
                await self.connection.login(
                    self.options.username.get_secret_value(),
                    self.options.password.get_secret_value(),
                )

            return True
        except OSError:
            if not self.options.fail_silently:
                raise

    async def close(self):
        if self.connection is None:  # pragma: no cover
            return

        try:
            await self.connection.quit()
        except Exception:
            if self.options.fail_silently:
                return
            raise
        finally:
            self.connection = None

    async def send(self) -> bool:
        try:
            await self.connection.send_message(self.options.email)
        except Exception:
            self.options.response = SmtpResponse(success=False)
            if self.options.fail_silently:
                return False
            raise

        self.options.response = SmtpResponse(success=True)
        return True



from accentnotifications.notifications.base import BaseNotification


class NotificationManager:
    async def send(self, options: BaseNotification):
        async with options.backend(options) as backend:
            return await backend.send()
